#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdbool.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>

#define Fflush() while(getchar() != '\n')

bool fehler = false;

typedef struct {
	char name[50];
	char telnr[20];
} satztyp;

bool satzgueltig(int snr, FILE* fp);
int anzahlsaetze(FILE* fp);
bool lesesatz(int snr, FILE* fp, satztyp* actsatz);
bool bestaetigeSatz(int snr, FILE* fp, satztyp* actsatz);
bool bestaetigeTausch(int snr1, int snr2, FILE* fp, satztyp* actsatz1, satztyp* actsatz2);

int main(int argc, char** argv){
	char* fname = (char*)malloc(100);
	FILE* fp = NULL;
	bool ende = false;
	int abfrage, i, anz;
	int snr1, snr2;
	satztyp actsatz, actsatz2;
	char cabfrage='j', zk[50];
	
	int rc1, rc2, fd, err = 0;
	struct flock lock1, lock2;
	
	do{
		if(argc==1 || cabfrage!='j'){
			printf("\nEinzulesende Datei angeben: ");
			scanf("%s", fname); Fflush();
		}
		else strncpy(fname, argv[1], 100);			// Filenamen lesen
	
		fp=fopen(fname,"rb");					// File existent?
		if(!fp){						// Datei existiert nicht
			perror("Fehler");
			printf("Datei neu anlegen (j/n, q=quit)?: ");
			cabfrage=getchar(); Fflush();
			if(cabfrage=='J' || cabfrage=='j')
				fp=fopen(fname,"ab");			// File anlegen (Append-Modus)
			else if(cabfrage=='q' || cabfrage=='Q')
				return 1;				// Quit
		}
		else cabfrage='j';					// Datei existiert, daher weiter
	}while(cabfrage!='j' && cabfrage!='J');
	fclose(fp);
	
	do{
		printf("HAUPTMENUE:\n===========\n\n");
		printf("1) Neuen Datensatz eingeben\n");
		printf("2) Datensaetze ausgeben\n");
		printf("3) Datensatz aendern\n");
		printf("4) Ende\n");
		printf("> ");
		scanf("%d", &abfrage);
		Fflush();
		
		printf("\n");
		switch(abfrage){
		case 1: // neuen Datensatz einlesen
			if((fp=fopen(fname, "ab")) == NULL){
				perror("fopen");
				return 2;
			}
			// nur ab Fileende locken
			fd = fileno(fp);
			lock1.l_type = F_WRLCK;
			lock1.l_whence = SEEK_END;
			lock1.l_start = 0;
			lock1.l_len = 0;
			printf("Warte auf Freigabe des Fileendes...\n");
			if(fcntl(fd, F_SETLKW, &lock1) < 0){
				perror("fnctl");
				return 4;
			}
			
			printf("Datensatz eingeben:\n");
			printf("Name: ");
			fgets(actsatz.name,50,stdin); 
			actsatz.name[strlen(actsatz.name)-1]='\0';
			printf("Tel.-Nr.: ");
			fgets(actsatz.telnr,20,stdin); 
			actsatz.telnr[strlen(actsatz.telnr)-1]='\0';
			
			if(!fwrite(&actsatz,sizeof(satztyp),1,fp)){
				fclose(fp);
				perror("fwrite");
				return 3;
			}
			lock1.l_type = F_UNLCK;
			if(fcntl(fd, F_SETLKW, &lock1) < 0){
				perror("fnctl");
				return 8;
			}
			printf("Fileende wieder freigegeben\n");
			fclose(fp);
			break;
		case 2: // Datensaetze ausgeben
			printf("Gespeicherte Datensaetze:\n\n");
			if((fp=fopen(fname, "rb")) == NULL){
				perror("fopen");
				return 2;
			}
			fd = fileno(fp);
			
			lock1.l_whence = 0;
			
			if((anz = anzahlsaetze(fp)) < 0){
				return 3;
			}
			if(anz == 0){
				printf("Keine Datensaetze im File vorhanden...\n");
			}
			else{
				for(i=1; i<=anz; i++){
					lock1.l_type = F_RDLCK;
					lock1.l_start = (i-1)*sizeof(satztyp);
					lock1.l_len = sizeof(satztyp);
					printf("Warte auf Freigabe des Datensatzes %d\n", i);
					if(fcntl(fd, F_SETLKW, &lock1) < 0){
						perror("fcntl");
						return 3;
					}
					printf("Datensatz %d gelockt (shared)\n", i);
					
					if((rc1=fread(&actsatz,sizeof(satztyp),1,fp))){
						printf("\nDatensatz Nr. %d:\n",i);
						printf("Name:     %s\n",actsatz.name);
						printf("Tel.-Nr.: %s\n",actsatz.telnr);
					}
					else{
						printf("rc=%d\n", rc1);
						printf("Satz %d nicht gelesen\n", i);
					}
					if(ferror(fp)){
						fclose(fp);
						perror("fread");
						return 4;
					}
				
					lock1.l_type = F_UNLCK;
					if(fcntl(fd, F_SETLKW, &lock1) < 0){
						perror("fcntl");
						return 5;
					}
					printf("Datensatz %d wieder freigegeben\n", i);
				}
			}
			fclose(fp);
			break;
		case 3: // Datensaetze aendern
			printf("Bitte waehlen:\n");
			printf("1) Aenderung mit Satznummer\n");
			printf("2) Austausch zweier Datensaetze\n");
			printf("> ");
			scanf("%d", &abfrage);
			Fflush();
			if((fp=fopen(fname, "rb+")) == NULL){ // Oeffnen im Update-Modus
				perror("fopen");
				return 2;
			}
			fd = fileno(fp);
			if(abfrage == 1){ // Aenderung anhand Satznummer
				printf("Satznummer: ");
				scanf("%d", &snr1);
				Fflush();
				
				lock1.l_type = F_WRLCK;
				lock1.l_whence = 0;
				lock1.l_start = (snr1-1)*sizeof(satztyp);
				lock1.l_len = sizeof(satztyp);
				printf("Warte auf Freigabe des Datensatzes %d\n", snr1);
				if(fcntl(fd, F_SETLKW, &lock1) < 0){
					perror("fnctl");
					return 4;
				}
				printf("Datensatz %d gelockt (exklusiv)\n", snr1);
				
				if(!bestaetigeSatz(snr1, fp, &actsatz)){
					if(fehler){
						fclose(fp);
				 		return 5;
					}
				}
				else{
					printf("Name eingeben: ");
					fgets(zk,50,stdin); zk[strlen(zk)-1]='\0';
					if(strcmp(zk,"") != 0) // nicht aendern bei leerer Eingabe
						strncpy(actsatz.name,zk,50);
					printf("Tel.-Nr. eingeben: ");
					fgets(zk,20,stdin); zk[strlen(zk)-1]='\0';
					if(strcmp(zk,"") != 0)
						strncpy(actsatz.telnr,zk,20);
	
					if(fseek(fp,(snr1-1)*sizeof(satztyp),SEEK_SET) < 0){ // zu Datensatz snr1 spulen
						perror("fseek");
						return 6;
					}
					if(!fwrite(&actsatz,sizeof(satztyp),1,fp)){
						fclose(fp);
						perror("fwrite");
						return 7;
					}
				}
				lock1.l_type = F_UNLCK;
				if(fcntl(fd, F_SETLKW, &lock1) < 0){
					perror("fnctl");
					return 8;
				}
				printf("Datensatz %d wieder freigegeben\n", snr1);

			}else{ // Austausch zweier Datensaetze
				printf("Nummer des 1. Datensatzes: ");
				scanf("%d", &snr1);
				Fflush();
				printf("Nummer des 2. Datensatzes: ");
				scanf("%d", &snr2);
				Fflush();
				
				if(!satzgueltig(snr1, fp)){
					printf("Satz %d ist nicht gueltig...\n", snr1);
					break;
				}
				if(!satzgueltig(snr2, fp)){
					printf("Satz %d ist nicht gueltig...\n", snr2);
					break;
				}
				
				lock1.l_start = (snr1-1)*sizeof(satztyp);
				lock2.l_start = (snr2-1)*sizeof(satztyp);
				lock1.l_len = lock2.l_len = sizeof(satztyp);
				lock1.l_whence = lock2.l_whence = 0;
				// fhre aus, solange Deadlock-Gefahr erkannt
				do{
					// exklusiv locken
					lock1.l_type = lock2.l_type = F_WRLCK;
					printf("Warte auf Freigabe des Datensatzes %d\n", snr1);
					// locke Record snr1
					if((rc1 = fcntl(fd, F_SETLKW, &lock1)) < 0){
						perror("fcntl1");
						return 6;
					}
					printf("Datensatz %d gelockt (exklusiv)\n", snr1);
					printf("Warte auf Freigabe des Datensatzes %d\n", snr2);
					// locke Record snr2
					if((rc2 = fcntl(fd, F_SETLKW, &lock2)) < 0){
						// wenn Deadlock-Gefahr, dann snr1 wieder freigeben
						if(errno == EDEADLK){
							err = errno;
							lock1.l_type = F_UNLCK;
							fcntl(fd, F_SETLKW, &lock1);
							sleep(1);
						}else{
							perror("fcntl2");
							return 7;
						}
					}else{
						printf("Datensatz %d gelockt (exklusiv)\n", snr2);
					}
				}while(rc2<0 && err==EDEADLK);
				
				if(!bestaetigeTausch(snr1, snr2, fp, &actsatz, &actsatz2)){
					if(fehler){
						fclose(fp);
						return 8;
					}
				}else{ // tausche Saetze aus
					if(fseek(fp,(snr2-1)*sizeof(satztyp),SEEK_SET) < 0){ // zu Datensatz snr2 spulen
						fclose(fp);
						perror("fseek");
						return 12;
					}
					if(!fwrite(&actsatz,sizeof(satztyp),1,fp)){ // Satz snr1 auf snr2 schreiben
						fclose(fp);
						perror("fwrite");
						return 11;
					}
					if(fseek(fp,(snr1-1)*sizeof(satztyp),SEEK_SET) < 0){ // zu Datensatz snr1 spulen
						fclose(fp);
						perror("fseek");
						return 12;
					}
					if(!fwrite(&actsatz2,sizeof(satztyp),1,fp)){ // Satz snr2 auf snr1 schreiben
						fclose(fp);
						perror("fwrite");
						return 13;
					}
				}
				
				lock1.l_type = lock2.l_type = F_UNLCK;
				if(fcntl(fd, F_SETLKW, &lock1) < 0){
					perror("fcntl");
					return 14;
				}
				printf("Datensatz %d wieder freigegeben\n", snr1);
				if(fcntl(fd, F_SETLKW, &lock2) < 0){
					perror("fcntl");
					return 15;
				}
				printf("Datensatz %d wieder freigegeben\n", snr2);
			}
			fclose(fp);
			break;
		default:
			ende = true;
			break;
		}
		printf("\n");
	}while(!ende);
	
	fclose(fp);	
	return 0;
}

// Besttigen des ndern von Satz snr
bool bestaetigeSatz(int snr, FILE* fp, satztyp* actsatz){
	char abfrage;
	
	if(!lesesatz(snr, fp, actsatz))
		return false;
	
	printf("\nDatensatz Nr. %d:\n", snr);
	printf("Name:     %s\n",actsatz->name);
	printf("Tel.-Nr.: %s\n",actsatz->telnr);
	printf("=> Soll dieser Datensatz wirklich geaendert werden (j/n)?: ");
	abfrage=getchar(); Fflush();
	if(abfrage!='j' && abfrage!='J'){
		return false;
	}
	
	return true;
}

// Besttigen des Tauschs von snr1 und snr2
bool bestaetigeTausch(int snr1, int snr2, FILE* fp, satztyp* actsatz1, satztyp* actsatz2){
	char abfrage;
	
	if(!lesesatz(snr1, fp, actsatz1))
		return false;
	if(!lesesatz(snr2, fp, actsatz2))
		return false;
	
	printf("\nDatensatz Nr. %d:\n", snr1);
	printf("Name:     %s\n",actsatz1->name);
	printf("Tel.-Nr.: %s\n",actsatz1->telnr);
	printf("\nDatensatz Nr. %d:\n", snr2);
	printf("Name:     %s\n",actsatz2->name);
	printf("Tel.-Nr.: %s\n",actsatz2->telnr);
	printf("=> Sollen diese Datensaetze wirklich getauscht werden (j/n)?: ");
	abfrage=getchar(); Fflush();
	if(abfrage!='j' && abfrage!='J'){
		return false;
	}
	
	return true;
}

// prfe Gltigkeit von Satz snr
bool satzgueltig(int snr, FILE* fp){
	int i = anzahlsaetze(fp);
	
	if(snr<=0 || snr>i || fehler){
		return false;
	}
	
	return true;
}

// Lies die Anzahl der Saetze im File
int anzahlsaetze(FILE* fp){
	int i;
	
	if(fseek(fp,0,SEEK_END) < 0){ // Dateiende lesen
		perror("fseek");
		fehler = true;
		return -1;
	}
	i = ftell(fp); // Bytes bis zur akt. Filepos
	if(fseek(fp,0,SEEK_SET) < 0){ // Dateianfang lesen
		perror("fseek");
		fehler = true;
		return -1;
	}

	return i/sizeof(satztyp);
}

bool lesesatz(int snr, FILE* fp, satztyp* actsatz){
	if(fseek(fp,(snr-1)*sizeof(satztyp),SEEK_SET) < 0){ // zu Datensatz snr spulen
		perror("fseek");
		fehler = true;
		return false;
	}
	if(!fread(actsatz,sizeof(satztyp),1,fp)){ // Datensatz lesen
		perror("fread");
		fehler = true;
		return false;
	}
	
	return true;
}
